#include "XmlFile.h"
#include <ctype.h>

static ulong new_(XmlFile **p_this);

static void element_callback_(char *token, void *x);


void xml_file_new(XmlFile **p_this, char *filename, char *mode) {
  XmlFile *this;

  new_(p_this);
  this = *p_this;

  file_new(&this->_file, filename, mode);
}

void xml_file_new_wrap_existing(XmlFile **p_this, File *file) {
  XmlFile *this;

  new_(p_this);
  this = *p_this;

  file_rewind(file);
  this->_file = file;
  this->wrapped = true;
}

void xml_file_new_memory_read_current_element(
    XmlFile **p_this,
    XmlFile *other) {
  XmlFile *this;

  if (!other->isStartTag) {
    throw("Must call after reading a start tag");
  }

  xml_file_new(p_this, null, null);
  this = *p_this;

  xml_file_skip_to_end_tag(other, other->_tagName, element_callback_, this);
  file_rewind(this->_file);
}

void xml_file_new_memory_read_element(
    XmlFile **p_this,
    XmlFile *other,
    char *element_name) {
  xml_file_read_start_tag(other, element_name);
  xml_file_new_memory_read_current_element(p_this, other);
}

void xml_file_delete(XmlFile **p_this) {
  XmlFile *this;

  if ((this = *p_this) == null) {
    return;
  }

  deletev(&this->_attributes);
  deletev(&this->_tagName);
  deletev(&this->_token);

  if (!this->wrapped) {
    file_delete(&this->_file);
  }

  deleteo(p_this);
}

void xml_file_read_token(XmlFile *this) {
  if (!xml_file_try_read_token(this)) {
    throw("Could not read token");
  }
}

bool xml_file_try_read_token(XmlFile *this) {
  bool is_tag;
  bool start;
  byte read;
  char c;
  char *p;
  char *pe;
  ulong mx;
  ulong mx_new;
  File *file;

  file = this->_file;
  p = this->_token;
  mx = this->maxTokenChars;
  pe = p + mx;
  start = true;
  is_tag = false;

  for (;;) {
    if (start && this->readTagStart) {
      c = '<';
      this->readTagStart = false;
    } else {
      if (!file_try_read_byte(file, &read)) {
        if (start || !is_tag) {
          for (p--; p >= this->_token && *p == ' '; p--) {
          }

          *(++p) = '\0';
          return p > this->_token;
        } else {
          throw("File ended in the middle of a tag");
        }
      }

      c = (char)read;

      if (isspace(c)) {
        c = ' ';
      }
    }

    if (c == ' ' && (start || *(p - 1) == ' ')) {
      continue;
    }

    *p = c;

    if (++p == pe) {
      mx_new = mx << 1;
      renewup(&this->_token, mx, mx_new);
      p = this->_token + mx;
      pe = this->_token + mx_new;
      mx = mx_new;
      this->maxTokenChars = mx;
    }

    if (start) {
      if (c == '>') {
        throw("Token started with '>'");
      }

      is_tag = c == '<';
      this->isTag = is_tag;
      start = false;
    } else {
      if (c == '<') {
        if (is_tag) {
          throw("Nested tags");
        } else {
          this->readTagStart = true;
          *--p = '\0';
          return true;
        }
      } else if (c == '>') {
        if (!is_tag) {
          throw("Got unmatched '>' in content");
        } else {
          *p = '\0';
          return true;
        }
      }
    }
  }
}

void xml_file_parse_tag(XmlFile *this) {
  bool quoted;
  char c;
  char *p;
  char *q;
  char *qe;
  ulong mx;
  ulong mx_new;

  if (!this->isTag) {
    throw("Not a tag");
  }

  this->isStartTag = false;
  this->isEndTag = false;
  this->isEmptyElementTag = false;

  p = this->_token;

  if (*p != '<') {
    throw("Impossible: Expected '<' but got '%c'", *p);
  }

  p++;

  for (; (c = *p) != '\0' && isspace(c); p++) {
  }

  if (*p == '/') {
    this->isEndTag = true;
    p++;
  }

  q = this->_tagName;
  mx = this->maxTagNameChars;
  qe = q + mx;

  for (; (c = *p) != '\0' && !isspace(c) && c != '/' && c != '>'; p++) {
    *q = c;

    if (++q == qe) {
      mx_new = mx << 1;
      renewup(&this->_tagName, mx, mx_new);
      q = this->_tagName + mx;
      qe = this->_tagName + mx_new;
      mx = mx_new;
      this->maxTagNameChars = mx;
    }
  }

  *q = '\0';

  for (; (c = *p) != '\0' && isspace(c); p++) {
  }

  if (c == '\0') {
    throw("Impossible: String ended prematurely");
  }

  if (c == '/') {
    this->isEmptyElementTag = true;

    for (; (c = *++p) != '\0' && isspace(c); p++) {
    }

    if (c != '>') {
      throw("Impossible: Expected '>' but got '%c'", c);
    }
  } else {
    q = this->_attributes;
    mx = this->maxAttributesChars;
    qe = q + mx;
    quoted = false;

    for (; (c = *p) != '\0'; p++) {
      if (!quoted && (c == '/' || c == '>')) {
        break;
      }

      *q = c;

      if (c == '"') {
        quoted = !quoted;
      }

      if (++q == qe) {
        mx_new = mx << 1;
        renewup(&this->_attributes, mx, mx_new);
        q = this->_attributes + mx;
        qe = this->_attributes + mx_new;
        mx = mx_new;
        this->maxAttributesChars = mx;
      }
    }

    for (q--; q >= this->_attributes && *q == ' '; q--) {
    }

    *(++q) = '\0';

    if (c == '\0') {
      throw("Impossible: String ended prematurely");
    }

    if (c == '>') {
      if (!this->isEndTag) {
        this->isStartTag = true;
      }
    } else {
      if (c != '/') {
        throw("Impossible: Expected '/' but got '%c'", c);
      }

      this->isEmptyElementTag = true;

      for (; (c = *++p) != '\0' && isspace(c); p++) {
      }

      if (c != '>') {
        throw("Impossible: Expected '>' but got '%c'", c);
      }
    }
  }
}

void xml_file_skip_to_end_tag(
    XmlFile *this,
    char *tag_name,
    void (*callback)(char *token, void *x),
    void *x) {
  char *_nested_tag_name = null;
  char *_tag_name_copy = null;

  string_new_copy(&_tag_name_copy, tag_name);

  for (;;) {
    xml_file_read_token(this);

    if (this->isTag) {
      xml_file_parse_tag(this);

      if (this->isEndTag) {
        if (!string_equals(this->_tagName, _tag_name_copy)) {
          throw("Found end tag for '%s' while skipping to end tag for '%s'",
              this->_tagName,
              tag_name);
        }

        string_delete(&_tag_name_copy);
        return;
      }

      if (callback != null) {
        callback(this->_token, x);
      }

      if (this->isStartTag) {
        string_new_copy(&_nested_tag_name, this->_tagName);
        xml_file_skip_to_end_tag(this, _nested_tag_name, callback, x);
        string_delete(&_nested_tag_name);

        if (callback != null) {
          callback(this->_token, x);
        }
      }
    } else {
      if (callback != null) {
        callback(this->_token, x);
      }
    }
  }
}

void xml_file_skip_to_element(
    XmlFile *this,
    char *tag_name,
    void (*callback)(char *token, void *x),
    void *x) {
  if (!xml_file_try_skip_to_element(this, tag_name, callback, x)) {
    throw("Could not skip to element '%s'", tag_name);
  }
}

bool xml_file_try_skip_to_element(
    XmlFile *this,
    char *tag_name,
    void (*callback)(char *token, void *x),
    void *x) {
  char *_nested_tag_name = null;

  for (;;) {
    if (!xml_file_try_read_token(this)) {
      return false;
    }

    if (this->isTag) {
      xml_file_parse_tag(this);

      if (this->isEndTag) {
        return false;
      }

      if (string_equals(this->_tagName, tag_name)) {
        return true;
      }

      if (callback != null) {
        callback(this->_token, x);
      }

      if (this->isStartTag) {
        string_new_copy(&_nested_tag_name, this->_tagName);
        xml_file_skip_to_end_tag(this, _nested_tag_name, callback, x);
        string_delete(&_nested_tag_name);

        if (callback != null) {
          callback(this->_token, x);
        }
      }

    } else {
      if (callback != null) {
        callback(this->_token, x);
      }
    }
  }
}

void xml_file_read_tag(XmlFile *this, char *tag_name) {
  xml_file_read_token(this);
  xml_file_parse_tag(this);

  if (!string_equals(this->_tagName, tag_name)) {
    throw("Tag name '%s' is not expected ('%s')", this->_tagName, tag_name);
  }
}

void xml_file_read_tag_attributes(
    XmlFile *this,
    char *tag_name,
    char *attributes) {
  xml_file_read_tag(this, tag_name);

  if (!string_equals(this->_attributes, attributes)) {
    throw("Attributes string '%s' is not expected ('%s')",
        this->_attributes,
        attributes);
  }
}

void xml_file_read_start_tag(XmlFile *this, char *tag_name) {
  xml_file_read_tag(this, tag_name);

  if (!this->isStartTag) {
    throw("Expected tag '%s' to be a start tag but it was not", tag_name);
  }
}

void xml_file_read_start_tag_attributes(
    XmlFile *this,
    char *tag_name,
    char *attributes) {
  xml_file_read_tag_attributes(this, tag_name, attributes);

  if (!this->isStartTag) {
    throw("Expected tag '%s' with attributes '%s' to be a start tag "
        "but it was not",
        tag_name,
        attributes);
  }
}

void xml_file_skip_element(XmlFile *this, char *element_name) {
  xml_file_read_start_tag(this, element_name);
  xml_file_skip_to_end_tag(this, element_name, null, null);
}

void xml_file_read_element(
    XmlFile *this,
    char *element_name,
    char **pcontents) {
  char *contents;
  XmlFile *_el = null;

  xml_file_new_memory_read_element(&_el, this, element_name);

  string_new_memory_file_contents(pcontents, _el->_file->_mf);
  contents = *pcontents;

  string_trim_trailing_whitespace(contents);
  xml_file_delete(&_el);
}

void xml_file_print(XmlFile *this, Printer *p) {
  sprint(p, "XmlFile current token");
  rprint(p, "Token: '%s'", this->_token);
  rprint(p, "Is tag: %s", bool_string(this->isTag));

  if (this->isTag) {
    xml_file_parse_tag(this);
    rprint(p, "Is start tag: %s", bool_string(this->isStartTag));
    rprint(p, "Is end tag: %s", bool_string(this->isEndTag));
    rprint(p,
        "Is empty element tag: %s",
        bool_string(this->isEmptyElementTag));
    rprint(p, "Tag name: '%s'", this->_tagName);
    rprint(p, "Attributes: '%s'", this->_attributes);
  }

  fprint(p);
}

void xml_file_write_token(XmlFile *this, char *fmt, ...) {
  char *_s = null;
  va_list ap1;
  va_list ap2;

  va_start(ap1, fmt);
  va_start(ap2, fmt);
  string_new_vf(&_s, fmt, ap1, ap2);
  va_end(ap1);
  va_end(ap2);

  file_write_string(this->_file, _s);
  string_delete(&_s);
}


static ulong new_(XmlFile **p_this) {
  XmlFile *this;

  this = newo(p_this);

  this->maxTokenChars = 1;
  new(&this->_token, 1);

  this->maxTagNameChars = 1;
  new(&this->_tagName, 1);

  this->maxAttributesChars = 1;
  new(&this->_attributes, 1);

  return 0;
}

static void element_callback_(char *token, void *x) {
  XmlFile *this;

  this = x;

  xml_file_write_token(this, token);
}
