#include "magic_double_voxmap.h"

DoubleVoxmap *double_voxmap_new_resize(
    DoubleVoxmap **p_this,
    DoubleVoxmap *other,
    ulong width,
    ulong height,
    ulong depth,
    double k,
    char *method,
    byte a,
    byte v) {
  byte num_kernel_pairs;
  Kernel **_x_kernels = null;
  Kernel **_y_kernels = null;
  Kernel **_z_kernels = null;

  kernels_new_resize(
      &_x_kernels,
      other->width,
      width,
      false,
      k,
      method,
      a,
      v,
      &num_kernel_pairs);

  kernels_new_resize(
      &_y_kernels,
      other->height,
      height,
      false,
      k,
      method,
      a,
      v,
      null);

  kernels_new_resize(
      &_z_kernels,
      other->depth,
      depth,
      false,
      k,
      method,
      a,
      v,
      null);

  double_voxmap_new_apply_multiple_xyz_kernels(
    p_this,
    other,
    num_kernel_pairs,
    _x_kernels,
    _y_kernels,
    _z_kernels);

  delete(&_z_kernels, num_kernel_pairs, kernel_delete);
  delete(&_y_kernels, num_kernel_pairs, kernel_delete);
  delete(&_x_kernels, num_kernel_pairs, kernel_delete);

  return *p_this;
}

DoubleVoxmap *double_voxmap_new_blur(
    DoubleVoxmap **p_this,
    DoubleVoxmap *other,
    double blur) {
  Kernel *_x_kernel = null;
  Kernel *_y_kernel = null;
  Kernel *_z_kernel = null;

  kernel_new_blur(&_x_kernel, other->width, false, blur);
  kernel_new_blur(&_y_kernel, other->height, false, blur);
  kernel_new_blur(&_z_kernel, other->depth, false, blur);

  double_voxmap_new_apply_xyz_kernels(
      p_this,
      other,
      _x_kernel,
      _y_kernel,
      _z_kernel);

  kernel_delete(&_z_kernel);
  kernel_delete(&_y_kernel);
  kernel_delete(&_x_kernel);

  return *p_this;
}

DoubleVoxmap *double_voxmap_new_sharpen(
    DoubleVoxmap **p_this,
    DoubleVoxmap *other,
    double sharpening) {
  Kernel *_x_kernel = null;
  Kernel *_y_kernel = null;
  Kernel *_z_kernel = null;

  kernel_new_sharpen(&_x_kernel, other->width, false, sharpening);
  kernel_new_sharpen(&_y_kernel, other->height, false, sharpening);
  kernel_new_sharpen(&_z_kernel, other->depth, false, sharpening);

  double_voxmap_new_apply_xyz_kernels(
      p_this,
      other,
      _x_kernel,
      _y_kernel,
      _z_kernel);

  kernel_delete(&_z_kernel);
  kernel_delete(&_y_kernel);
  kernel_delete(&_x_kernel);

  return *p_this;
}
