#include "DoubleVoxmapTest.h"

START

TEST(DoubleVoxmap, New) {
  ulong i;
  ulong j;
  ulong k;
  DoubleVoxmap *_dv = null;

  double_voxmap_new(&_dv, 3, 4, 5);
  EXPECT_ULEQ(3, _dv->width);
  EXPECT_ULEQ(4, _dv->height);
  EXPECT_ULEQ(5, _dv->depth);
  EXPECT_ULEQ(60, _dv->numVoxels);

  for (i = 0; i < 3; i++) {
    for (j = 0; j < 4; j++) {
      for (k = 0; k < 5; k++) {
        EXPECT_DEQ(0, double_voxmap_get(_dv, i, j, k));
      }
    }
  }

  double_voxmap_delete(&_dv);
}

TEST(DoubleVoxmap, Print) {
  char *_s = null;
  DoubleVoxmap *_ = null;

  double_voxmap_new(&_, 1, 2, 3);
  double_voxmap_set(_, 0, 0, 1, 0.6);
  double_voxmap_set(_, 0, 1, 2, 0.3);
  double_voxmap_set(_, 0, 0, 0, 0.5);
  aprint(&_s, 0, double_voxmap, _);
  EXPECT_STREQ(
      "[DoubleVoxmap]\n. Width: 1\n. Height: 2\n. Depth: 3\n. [Contents]\n"
      ". . [Slice 0]\n. . . [Row 0]\n. . . . [0] 0.5\n. . . [/Row 0]\n"
      ". . . [Row 1]\n. . . . [0] 0\n. . . [/Row 1]\n. . [/Slice 0]\n"
      ". . [Slice 1]\n. . . [Row 0]\n. . . . [0] 0.6\n. . . [/Row 0]\n"
      ". . . [Row 1]\n. . . . [0] 0\n. . . [/Row 1]\n. . [/Slice 1]\n"
      ". . [Slice 2]\n. . . [Row 0]\n. . . . [0] 0\n. . . [/Row 0]\n"
      ". . . [Row 1]\n. . . . [0] 0.3\n. . . [/Row 1]\n. . [/Slice 2]\n"
      ". [/Contents]\n[/DoubleVoxmap]\n",
      _s);
  string_delete(&_s);

  double_voxmap_delete(&_);
}

TEST(DoubleVoxmap, GetSet) {
  ulong i;
  ulong j;
  ulong k;
  double e;
  DoubleVoxmap *_dv = null;

  double_voxmap_new(&_dv, 2, 3, 4);
  double_voxmap_set(_dv, 0, 0, 0, 1.5);
  double_voxmap_set(_dv, 1, 2, 3, 2.6);
  double_voxmap_set(_dv, 0, 0, 1, -1.3);
  double_voxmap_set(_dv, 0, 2, 3, 1000);

  for (i = 0; i < 2; i++) {
    for (j = 0; j < 3; j++) {
      for (k = 0; k < 4; k++) {
        if (i == 0 && j == 0 && k == 0) {
          e = 1.5;
        } else if (i == 1 && j == 2 && k == 3) {
          e = 2.6;
        } else if (i == 0 && j == 0 && k == 1) {
          e = -1.3;
        } else if (i == 0 && j == 2 && k == 3) {
          e = 1000;
        } else {
          e = 0;
        }

        EXPECT_DEQ(e, double_voxmap_get(_dv, i, j, k));
      }
    }
  }

  double_voxmap_delete(&_dv);
}

TEST(DoubleVoxmap, AddSubtractMaxAbsDiff) {
  ulong i;
  ulong j;
  ulong k;
  double x;
  double y;
  DoubleVoxmap *_a = null;
  DoubleVoxmap *_b = null;
  DoubleVoxmap *_sum = null;
  DoubleVoxmap *_diff = null;

  double_voxmap_new(&_a, 3, 4, 4);
  double_voxmap_new(&_b, 3, 4, 4);

  for (i = 0; i < 3; i++) {
    for (j = 0; j < 4; j++) {
      for (k = 0; k < 4; k++) {
        _a->_voxels[i][j][k] = (i + 1.0) / (j + 1.5) + (k + 1.7) / 11.3;
        _b->_voxels[i][j][k] = (i - 3.4) / (j + 1.6) + (k + 1.9) / 3.14;
      }
    }
  }

  double_voxmap_new_add(&_sum, _a, _b);
  double_voxmap_new_subtract(&_diff, _a, _b);

  for (i = 0; i < 3; i++) {
    for (j = 0; j < 4; j++) {
      for (k = 0; k < 4; k++) {
        x = (i + 1.0) / (j + 1.5) + (k + 1.7) / 11.3;
        y = (i - 3.4) / (j + 1.6) + (k + 1.9) / 3.14;
        EXPECT_DEQ(x + y, _sum->_voxels[i][j][k]);
        EXPECT_DEQ(x - y, _diff->_voxels[i][j][k]);
      }
    }
  }

  double_voxmap_delete(&_diff);
  double_voxmap_delete(&_sum);

  EXPECT_NEAR(2.42035, double_voxmap_max_abs_diff(_a, _b), 1e-5);

  double_voxmap_delete(&_b);
  double_voxmap_delete(&_a);
}

FINISH
