#include "XmlFileTest.h"

START

TEST(XmlFile, Tokens) {
  File *_f = null;
  XmlFile *_xml = null;

  file_new_memory_copy_string(&_f, "");
  xml_file_new_wrap_existing(&_xml, _f);
  EXPECT_FALSE(xml_file_try_read_token(_xml));
  xml_file_delete(&_xml);
  file_delete(&_f);

  file_new_memory_copy_string(&_f, " ");
  xml_file_new_wrap_existing(&_xml, _f);
  EXPECT_FALSE(xml_file_try_read_token(_xml));
  xml_file_delete(&_xml);
  file_delete(&_f);

  file_new_memory_copy_string(&_f, " \r\t\f\v ");
  xml_file_new_wrap_existing(&_xml, _f);
  EXPECT_FALSE(xml_file_try_read_token(_xml));
  xml_file_delete(&_xml);
  file_delete(&_f);

  file_new_memory_copy_string(&_f, "<woof>");
  xml_file_new_wrap_existing(&_xml, _f);
  EXPECT_TRUE(xml_file_try_read_token(_xml));
  EXPECT_STREQ("<woof>", _xml->_token);
  EXPECT_TRUE(_xml->isTag);
  xml_file_parse_tag(_xml);
  EXPECT_TRUE(_xml->isStartTag);
  EXPECT_FALSE(_xml->isEndTag);
  EXPECT_FALSE(_xml->isEmptyElementTag);
  EXPECT_STREQ("woof", _xml->_tagName);
  EXPECT_STREQ("", _xml->_attributes);
  xml_file_delete(&_xml);
  file_delete(&_f);

  file_new_memory_copy_string(&_f, "<a b >");
  xml_file_new_wrap_existing(&_xml, _f);
  EXPECT_TRUE(xml_file_try_read_token(_xml));
  EXPECT_STREQ("<a b >", _xml->_token);
  EXPECT_TRUE(_xml->isTag);
  xml_file_parse_tag(_xml);
  EXPECT_TRUE(_xml->isStartTag);
  EXPECT_FALSE(_xml->isEndTag);
  EXPECT_FALSE(_xml->isEmptyElementTag);
  EXPECT_STREQ("a", _xml->_tagName);
  EXPECT_STREQ("b", _xml->_attributes);
  xml_file_delete(&_xml);
  file_delete(&_f);

  file_new_memory_copy_string(&_f, "< tag  dogger\ncat\n\n \r\f\v>");
  xml_file_new_wrap_existing(&_xml, _f);
  EXPECT_TRUE(xml_file_try_read_token(_xml));
  EXPECT_STREQ("< tag dogger cat >", _xml->_token);
  EXPECT_TRUE(_xml->isTag);
  xml_file_parse_tag(_xml);
  EXPECT_TRUE(_xml->isStartTag);
  EXPECT_FALSE(_xml->isEndTag);
  EXPECT_FALSE(_xml->isEmptyElementTag);
  EXPECT_STREQ("tag", _xml->_tagName);
  EXPECT_STREQ("dogger cat", _xml->_attributes);
  xml_file_delete(&_xml);
  file_delete(&_f);

  file_new_memory_copy_string(&_f, "<dog>cat");
  xml_file_new_wrap_existing(&_xml, _f);
  EXPECT_TRUE(xml_file_try_read_token(_xml));
  EXPECT_TRUE(xml_file_try_read_token(_xml));
  EXPECT_STREQ("cat", _xml->_token);
  EXPECT_FALSE(_xml->isTag);
  xml_file_delete(&_xml);
  file_delete(&_f);

  file_new_memory_copy_string(&_f, "<dog> \rcat \r ");
  xml_file_new_wrap_existing(&_xml, _f);
  EXPECT_TRUE(xml_file_try_read_token(_xml));
  EXPECT_TRUE(xml_file_try_read_token(_xml));
  EXPECT_STREQ("cat", _xml->_token);
  EXPECT_FALSE(_xml->isTag);
  xml_file_delete(&_xml);
  file_delete(&_f);

  file_new_memory_copy_string(&_f, "<dog>cat</dog>");
  xml_file_new_wrap_existing(&_xml, _f);
  EXPECT_TRUE(xml_file_try_read_token(_xml));
  EXPECT_TRUE(xml_file_try_read_token(_xml));
  EXPECT_TRUE(xml_file_try_read_token(_xml));
  EXPECT_STREQ("</dog>", _xml->_token);
  EXPECT_TRUE(_xml->isTag);
  xml_file_parse_tag(_xml);
  EXPECT_FALSE(_xml->isStartTag);
  EXPECT_TRUE(_xml->isEndTag);
  EXPECT_FALSE(_xml->isEmptyElementTag);
  EXPECT_STREQ("dog", _xml->_tagName);
  EXPECT_STREQ("", _xml->_attributes);
  xml_file_delete(&_xml);
  file_delete(&_f);

  file_new_memory_copy_string(&_f, "<woof/>");
  xml_file_new_wrap_existing(&_xml, _f);
  EXPECT_TRUE(xml_file_try_read_token(_xml));
  EXPECT_STREQ("<woof/>", _xml->_token);
  EXPECT_TRUE(_xml->isTag);
  xml_file_parse_tag(_xml);
  EXPECT_FALSE(_xml->isStartTag);
  EXPECT_FALSE(_xml->isEndTag);
  EXPECT_TRUE(_xml->isEmptyElementTag);
  EXPECT_STREQ("woof", _xml->_tagName);
  EXPECT_STREQ("", _xml->_attributes);
  xml_file_delete(&_xml);
  file_delete(&_f);
}

TEST(XmlFile, Tags) {
  File *_f = null;
  XmlFile *_xml = null;

  file_new_memory_copy_string(&_f, "<dog>");
  xml_file_new_wrap_existing(&_xml, _f);
  xml_file_read_tag(_xml, "dog");
  xml_file_delete(&_xml);
  file_delete(&_f);

  file_new_memory_copy_string(&_f, "<dog size=1>");
  xml_file_new_wrap_existing(&_xml, _f);
  xml_file_read_tag_attributes(_xml, "dog", "size=1");
  xml_file_delete(&_xml);
  file_delete(&_f);

  file_new_memory_copy_string(&_f, "< dog  name=\"Fido\"   >");
  xml_file_new_wrap_existing(&_xml, _f);
  xml_file_read_tag_attributes(_xml, "dog", "name=\"Fido\"");
  xml_file_delete(&_xml);
  file_delete(&_f);

  file_new_memory_copy_string(&_f, "<dog>");
  xml_file_new_wrap_existing(&_xml, _f);
  xml_file_read_start_tag(_xml, "dog");
  xml_file_delete(&_xml);
  file_delete(&_f);

  file_new_memory_copy_string(&_f, "<dog cat=1>");
  xml_file_new_wrap_existing(&_xml, _f);
  xml_file_read_start_tag_attributes(_xml, "dog", "cat=1");
  xml_file_delete(&_xml);
  file_delete(&_f);

  file_new_memory_copy_string(&_f, "<dog></dog>");
  xml_file_new_wrap_existing(&_xml, _f);
  xml_file_read_start_tag(_xml, "dog");
  xml_file_skip_to_end_tag(_xml, "dog", null, null);
  EXPECT_STREQ("</dog>", _xml->_token);
  xml_file_delete(&_xml);
  file_delete(&_f);
}

TEST(XmlFile, Elements) {
  char *_contents = null;
  File *_f = null;
  XmlFile *_el = null;
  XmlFile *_xml = null;

  file_new_memory_copy_string(&_f, "<dog>Woof</dog>");
  xml_file_new_wrap_existing(&_xml, _f);
  xml_file_read_start_tag(_xml, "dog");
  xml_file_new_memory_read_current_element(&_el, _xml);
  xml_file_read_token(_el);
  EXPECT_STREQ("Woof", _el->_token);
  xml_file_delete(&_el);
  xml_file_delete(&_xml);
  file_delete(&_f);

  file_new_memory_copy_string(&_f, "<dog>Woof</dog>");
  xml_file_new_wrap_existing(&_xml, _f);
  xml_file_new_memory_read_element(&_el, _xml, "dog");
  xml_file_read_token(_el);
  EXPECT_STREQ("Woof", _el->_token);
  xml_file_delete(&_el);
  xml_file_delete(&_xml);
  file_delete(&_f);

  file_new_memory_copy_string(&_f, "<dog><cat>Woof</cat></dog>");
  xml_file_new_wrap_existing(&_xml, _f);
  xml_file_new_memory_read_element(&_el, _xml, "dog");
  xml_file_read_token(_el);
  EXPECT_STREQ("<cat>", _el->_token);
  xml_file_read_token(_el);
  EXPECT_STREQ("Woof", _el->_token);
  xml_file_read_token(_el);
  EXPECT_STREQ("</cat>", _el->_token);
  EXPECT_FALSE(xml_file_try_read_token(_el));
  xml_file_delete(&_el);
  xml_file_delete(&_xml);
  file_delete(&_f);

  file_new_memory_copy_string(&_f, "<dog>Woof</dog>");
  xml_file_new_wrap_existing(&_xml, _f);
  xml_file_read_element(_xml, "dog", &_contents);
  EXPECT_STREQ("Woof", _contents);
  string_delete(&_contents);
  xml_file_delete(&_xml);
  file_delete(&_f);

  file_new_memory_copy_string(&_f, "<dog>Woof</dog><cat><m>iaow</m></cat>");
  xml_file_new_wrap_existing(&_xml, _f);
  xml_file_skip_element(_xml, "dog");
  xml_file_read_element(_xml, "cat", &_contents);
  EXPECT_STREQ("<m>iaow</m>", _contents);
  string_delete(&_contents);
  xml_file_delete(&_xml);
  file_delete(&_f);

  file_new_memory_copy_string(&_f, "<dog>Woof</dog><cat><m>iaow</m></cat>");
  xml_file_new_wrap_existing(&_xml, _f);
  xml_file_skip_to_element(_xml, "cat", null, null);
  xml_file_read_element(_xml, "m", &_contents);
  EXPECT_STREQ("iaow", _contents);
  string_delete(&_contents);
  xml_file_delete(&_xml);
  file_delete(&_f);

  file_new_memory_copy_string(&_f, "<dog>Woof</dog><cat><m>iaow</m></cat>");
  xml_file_new_wrap_existing(&_xml, _f);
  EXPECT_FALSE(xml_file_try_skip_to_element(_xml, "pig", null, null));
  xml_file_delete(&_xml);
  file_delete(&_f);
}

TEST(XmlFile, PrintCurrentToken) {
  char *_s = null;
  File *_f = null;
  XmlFile *_xml = null;

  file_new_memory_copy_string(&_f, "<dog size=12>");
  xml_file_new_wrap_existing(&_xml, _f);
  xml_file_read_start_tag(_xml, "dog");
  aprint(&_s, 0, xml_file, _xml);
  EXPECT_STREQ(
      "[XmlFile current token]\n. Token: '<dog size=12>'\n"
      ". Is tag: true\n. Is start tag: true\n. Is end tag: false\n"
      ". Is empty element tag: false\n. Tag name: 'dog'\n"
      ". Attributes: 'size=12'\n[/XmlFile current token]\n",
      _s);
  string_delete(&_s);
  xml_file_delete(&_xml);
  file_delete(&_f);
}

FINISH
