#include "date_test.h"

START

TEST(Date, Validate) {
  date_validate("1901-01-01");
  date_validate("2079-12-31");
  date_validate("1966-09-24");
  date_validate("1968-07-07");
}

TEST(Date, TryValidate) {
  EXPECT_FALSE(date_try_validate("2001-1-1"));
  EXPECT_FALSE(date_try_validate("2001/01/01"));
  EXPECT_FALSE(date_try_validate("2001- 1-01"));
  EXPECT_FALSE(date_try_validate("  99-01-01"));
  EXPECT_FALSE(date_try_validate("2001-1 -01"));
  EXPECT_FALSE(date_try_validate("2001-01- 1"));
  EXPECT_FALSE(date_try_validate("2001-01-1 "));
  EXPECT_FALSE(date_try_validate("1900-01-01"));
  EXPECT_FALSE(date_try_validate("2080-01-01"));
  EXPECT_FALSE(date_try_validate("2014-00-01"));
  EXPECT_FALSE(date_try_validate("2014-13-01"));
  EXPECT_FALSE(date_try_validate("2014-99-01"));
  EXPECT_FALSE(date_try_validate("2014-01-00"));
  EXPECT_FALSE(date_try_validate("2014-01-32"));
  EXPECT_FALSE(date_try_validate("2014-01-99"));
  EXPECT_FALSE(date_try_validate("2014-02-29"));
  EXPECT_TRUE(date_try_validate("2012-02-29"));
  EXPECT_FALSE(date_try_validate("2014-03-32"));
  EXPECT_FALSE(date_try_validate("2014-04-31"));
  EXPECT_FALSE(date_try_validate("2014-06-31"));
  EXPECT_FALSE(date_try_validate("2014-09-31"));
  EXPECT_FALSE(date_try_validate("2014-11-31"));
}

TEST(Date, Parse) {
  byte day;
  byte month;
  ushort year;

  date_parse("1966-09-24", &day, &month, &year);
  EXPECT_ULEQ(24, day);
  EXPECT_ULEQ(9, month);
  EXPECT_ULEQ(1966, year);

  date_parse("1901-01-01", &day, &month, &year);
  EXPECT_ULEQ(1, day);
  EXPECT_ULEQ(1, month);
  EXPECT_ULEQ(1901, year);

  date_parse("2079-12-31", &day, &month, &year);
  EXPECT_ULEQ(31, day);
  EXPECT_ULEQ(12, month);
  EXPECT_ULEQ(2079, year);
}

TEST(Date, TryParse) {
  byte day;
  byte month;
  ushort year;

  EXPECT_TRUE(date_try_parse("1966-09-24", &day, &month, &year));
  EXPECT_ULEQ(24, day);
  EXPECT_ULEQ(9, month);
  EXPECT_ULEQ(1966, year);

  EXPECT_TRUE(date_try_parse("1901-01-01", &day, &month, &year));
  EXPECT_ULEQ(1, day);
  EXPECT_ULEQ(1, month);
  EXPECT_ULEQ(1901, year);

  EXPECT_TRUE(date_try_parse("2079-12-31", &day, &month, &year));
  EXPECT_ULEQ(31, day);
  EXPECT_ULEQ(12, month);
  EXPECT_ULEQ(2079, year);

  EXPECT_FALSE(date_try_parse("19660924", &day, &month, &year));
  EXPECT_FALSE(date_try_parse("1966 09 24", &day, &month, &year));
  EXPECT_FALSE(date_try_parse("09/24/1966", &day, &month, &year));
  EXPECT_FALSE(date_try_parse(" 1966-09-24", &day, &month, &year));
  EXPECT_FALSE(date_try_parse("1966-09-24 ", &day, &month, &year));
}

TEST(Date, Day) {
  EXPECT_ULEQ(24, date_day("1966-09-24"));
}

TEST(Date, Month) {
  EXPECT_ULEQ(9, date_month("1966-09-24"));
}

TEST(Date, MonthShortname) {
  EXPECT_STREQ("Jan", date_month_shortname("1931-01-01"));
  EXPECT_STREQ("Feb", date_month_shortname("2001-02-12"));
  EXPECT_STREQ("Mar", date_month_shortname("1923-03-21"));
  EXPECT_STREQ("Apr", date_month_shortname("2022-04-01"));
  EXPECT_STREQ("May", date_month_shortname("2009-05-16"));
  EXPECT_STREQ("Jun", date_month_shortname("2001-06-30"));
  EXPECT_STREQ("Jul", date_month_shortname("1968-07-07"));
  EXPECT_STREQ("Aug", date_month_shortname("2015-08-11"));
  EXPECT_STREQ("Sep", date_month_shortname("1966-09-24"));
  EXPECT_STREQ("Oct", date_month_shortname("1979-10-09"));
  EXPECT_STREQ("Nov", date_month_shortname("2001-11-09"));
  EXPECT_STREQ("Dec", date_month_shortname("2001-12-25"));
}

TEST(Date, MonthShortnameFromValue) {
  byte a;

  a = 1;
  EXPECT_STREQ("Jan", date_month_shortname_from_value(a++));
  EXPECT_STREQ("Feb", date_month_shortname_from_value(a++));
  EXPECT_STREQ("Mar", date_month_shortname_from_value(a++));
  EXPECT_STREQ("Apr", date_month_shortname_from_value(a++));
  EXPECT_STREQ("May", date_month_shortname_from_value(a++));
  EXPECT_STREQ("Jun", date_month_shortname_from_value(a++));
  EXPECT_STREQ("Jul", date_month_shortname_from_value(a++));
  EXPECT_STREQ("Aug", date_month_shortname_from_value(a++));
  EXPECT_STREQ("Sep", date_month_shortname_from_value(a++));
  EXPECT_STREQ("Oct", date_month_shortname_from_value(a++));
  EXPECT_STREQ("Nov", date_month_shortname_from_value(a++));
  EXPECT_STREQ("Dec", date_month_shortname_from_value(a++));
}

TEST(Date, MonthLongname) {
  EXPECT_STREQ("January", date_month_longname("1931-01-01"));
  EXPECT_STREQ("February", date_month_longname("2001-02-12"));
  EXPECT_STREQ("March", date_month_longname("1923-03-21"));
  EXPECT_STREQ("April", date_month_longname("2022-04-01"));
  EXPECT_STREQ("May", date_month_longname("2009-05-16"));
  EXPECT_STREQ("June", date_month_longname("2001-06-30"));
  EXPECT_STREQ("July", date_month_longname("1968-07-07"));
  EXPECT_STREQ("August", date_month_longname("2015-08-11"));
  EXPECT_STREQ("September", date_month_longname("1966-09-24"));
  EXPECT_STREQ("October", date_month_longname("1979-10-09"));
  EXPECT_STREQ("November", date_month_longname("2001-11-09"));
  EXPECT_STREQ("December", date_month_longname("2001-12-25"));
}

TEST(Date, MonthLongnameFromValue) {
  byte a;

  a = 1;
  EXPECT_STREQ("January", date_month_longname_from_value(a++));
  EXPECT_STREQ("February", date_month_longname_from_value(a++));
  EXPECT_STREQ("March", date_month_longname_from_value(a++));
  EXPECT_STREQ("April", date_month_longname_from_value(a++));
  EXPECT_STREQ("May", date_month_longname_from_value(a++));
  EXPECT_STREQ("June", date_month_longname_from_value(a++));
  EXPECT_STREQ("July", date_month_longname_from_value(a++));
  EXPECT_STREQ("August", date_month_longname_from_value(a++));
  EXPECT_STREQ("September", date_month_longname_from_value(a++));
  EXPECT_STREQ("October", date_month_longname_from_value(a++));
  EXPECT_STREQ("November", date_month_longname_from_value(a++));
  EXPECT_STREQ("December", date_month_longname_from_value(a++));
}

TEST(Date, Min) {
  EXPECT_STREQ("1901-01-01", date_min());
}

TEST(Date, Max) {
  EXPECT_STREQ("2079-12-31", date_max());
}

TEST(Date, NewCopy) {
  char *_ = null;

  string_new_date_copy(&_, "1966-09-24");
  EXPECT_STREQ("1966-09-24", _);
  string_delete(&_);
}

TEST(Date, NewMin) {
  char *_ = null;

  string_new_date_min(&_);
  EXPECT_STREQ("1901-01-01", _);
  string_delete(&_);
}

TEST(Date, NewMax) {
  char *_ = null;

  string_new_date_max(&_);
  EXPECT_STREQ("2079-12-31", _);
  string_delete(&_);
}

TEST(Date, NewToday) {
  char *_ = null;
  time_t now;
  struct tm *t;
  Lock l = LOCK_INITIALIZER;

  lock(&l);

  now = time(null);
  t = localtime(&now);

  string_new_date_today(&_, false);
  EXPECT_ULEQ(t->tm_mday, date_day(_));
  EXPECT_ULEQ(t->tm_mon + 1, date_month(_));
  EXPECT_ULEQ(t->tm_year + 1900, date_year(_));
  string_delete(&_);

  now = time(null);
  t = gmtime(&now);

  string_new_date_today(&_, true);
  EXPECT_ULEQ(t->tm_mday, date_day(_));
  EXPECT_ULEQ(t->tm_mon + 1, date_month(_));
  EXPECT_ULEQ(t->tm_year + 1900, date_year(_));
  string_delete(&_);

  unlock(&l);
}

TEST(Date, NewYesterday) {
  char *_today = null;
  char *_yesterday = null;

  string_new_date_today(&_today, false);
  string_new_date_yesterday(&_yesterday, false);
  EXPECT_LEQ(1, date_subtract(_today, _yesterday));
  string_delete(&_yesterday);
  string_delete(&_today);

  string_new_date_today(&_today, true);
  string_new_date_yesterday(&_yesterday, true);
  EXPECT_LEQ(1, date_subtract(_today, _yesterday));
  string_delete(&_yesterday);
  string_delete(&_today);
}

TEST(Date, NewNext) {
  char *_ = null;

  string_new_date_next(&_, "1901-12-31");
  EXPECT_STREQ("1902-01-01", _);
  string_delete(&_);
}

TEST(Date, NewPrevious) {
  char *_ = null;

  string_new_date_previous(&_, "1902-01-01");
  EXPECT_STREQ("1901-12-31", _);
  string_delete(&_);
}

TEST(Date, RenewIncrement) {
  char *_ = null;

  for (
      string_new_date_min(&_);
      date_earlier(_, date_max());
      string_renew_date_increment(&_)) {
    EXPECT_TRUE(date_try_validate(_));
  }

  EXPECT_STREQ("2079-12-31", _);
  string_delete(&_);
}

TEST(Date, RenewDecrement) {
  char *_ = null;

  for (
      string_new_date_max(&_);
      date_later(_, date_min());
      string_renew_date_decrement(&_)) {
    EXPECT_TRUE(date_try_validate(_));
  }

  EXPECT_STREQ("1901-01-01", _);
  string_delete(&_);
}

TEST(Date, NewAdd) {
  char *_ = null;

  string_new_date_add(&_, "1979-09-24", 15);
  EXPECT_STREQ("1979-10-09", _);
  string_delete(&_);

  string_new_date_add(&_, "1979-10-09", -15);
  EXPECT_STREQ("1979-09-24", _);
  string_delete(&_);

  string_new_date_add(&_, "1966-09-24", 652);
  EXPECT_STREQ("1968-07-07", _);
  string_delete(&_);

  string_new_date_add(&_, "1968-07-07", -652);
  EXPECT_STREQ("1966-09-24", _);
  string_delete(&_);

}

TEST(Date, NewSubtract) {
  char *_ = null;

  string_new_date_subtract(&_, "1979-09-24", -15);
  EXPECT_STREQ("1979-10-09", _);
  string_delete(&_);

  string_new_date_subtract(&_, "1979-10-09", 15);
  EXPECT_STREQ("1979-09-24", _);
  string_delete(&_);

  string_new_date_subtract(&_, "1966-09-24", -652);
  EXPECT_STREQ("1968-07-07", _);
  string_delete(&_);

  string_new_date_subtract(&_, "1968-07-07", 652);
  EXPECT_STREQ("1966-09-24", _);
  string_delete(&_);
}

TEST(Date, RenewAdd) {
  char *_ = null;

  string_new_date_copy(&_, "1979-10-09");
  string_renew_date_add(&_, -15);
  EXPECT_STREQ("1979-09-24", _);
  string_renew_date_add(&_, 15);
  EXPECT_STREQ("1979-10-09", _);
  string_delete(&_);
}

TEST(Date, RenewSubtract) {
  char *_ = null;

  string_new_date_copy(&_, "1979-10-09");
  string_renew_date_subtract(&_, 15);
  EXPECT_STREQ("1979-09-24", _);
  string_renew_date_subtract(&_, -15);
  EXPECT_STREQ("1979-10-09", _);
  string_delete(&_);
}

TEST(Date, Subtract) {
  EXPECT_ULEQ(17356, date_subtract("2014-04-01", "1966-09-24"));
  EXPECT_LEQ(-17356, date_subtract("1966-09-24", "2014-04-01"));
  EXPECT_LEQ(1461, date_subtract("2014-09-24", "2010-09-24"));
  EXPECT_LEQ(0, date_subtract("2014-09-24", "2014-09-24"));
}

TEST(Date, Equals) {
  EXPECT_TRUE(date_equals("1966-09-24", "1966-09-24"));
  EXPECT_FALSE(date_equals("1966-09-24", "1967-09-24"));
  EXPECT_FALSE(date_equals("1966-09-24", "1966-08-24"));
  EXPECT_FALSE(date_equals("1966-09-24", "1966-09-25"));
}

TEST(Date, Earlier) {
  EXPECT_TRUE(date_earlier("1966-09-24", "1968-07-07"));
  EXPECT_FALSE(date_earlier("1968-07-07", "1966-09-24"));
  EXPECT_FALSE(date_earlier("1966-09-24", "1966-09-24"));
}

TEST(Date, EarlierOrEqual) {
  EXPECT_TRUE(date_earlier_or_equal("1966-09-24", "1968-07-07"));
  EXPECT_FALSE(date_earlier_or_equal("1968-07-07", "1966-09-24"));
  EXPECT_TRUE(date_earlier_or_equal("1966-09-24", "1966-09-24"));
}

TEST(Date, Later) {
  EXPECT_FALSE(date_later("1966-09-24", "1968-07-07"));
  EXPECT_TRUE(date_later("1968-07-07", "1966-09-24"));
  EXPECT_FALSE(date_later("1966-09-24", "1966-09-24"));
}

TEST(Date, LaterOrEqual) {
  EXPECT_FALSE(date_later_or_equal("1966-09-24", "1968-07-07"));
  EXPECT_TRUE(date_later_or_equal("1968-07-07", "1966-09-24"));
  EXPECT_TRUE(date_later_or_equal("1966-09-24", "1966-09-24"));
}

TEST(Date, Increment) {
  char *_ = null;

  string_new_date_copy(&_, "1966-12-31");
  date_increment(_);
  EXPECT_STREQ("1967-01-01", _);
  string_delete(&_);
}

TEST(Date, Decrement) {
  char *_ = null;

  string_new_date_copy(&_, "2000-01-01");
  date_decrement(_);
  EXPECT_STREQ("1999-12-31", _);
  string_delete(&_);
}

FINISH
