#include "directory_test_death.h"

START_DEATH

TEST(Directory, IsEmpty) {
  EXPECT_DEATH(
      directory_is_empty("/catdog/mouse"),
      "Directory '/catdog/mouse' does not exist");
}

TEST(Directory, Read) {
  char **_entries = null;

  EXPECT_DEATH(
      directory_read(
        "/catdog/mouse",
        null,
        null,
        false,
        false,
        &_entries),
      "Could not read the directory '/catdog/mouse'");
}

TEST(Directory, EnsureExists) {
  EXPECT_DEATH(
    directory_ensure_exists("/dev/stdout/dog"),
    "Could not ensure that directory '/dev/stdout/dog/' exists: \f");
}

TEST(Directory, Rename) {
  char *_dir1 = null;
  char *_dir2 = null;
  char *_e = null;
  TempFilename *_base = null;

  EXPECT_DEATH(
      directory_rename("cat", "dog/woof"),
      "Cannot rename directory 'cat' with 1 part to 'dog/woof' with 2 parts");
  EXPECT_DEATH(
      directory_rename("cat/miaow", "dog"),
      "Cannot rename directory 'cat/miaow' with 2 parts to 'dog' with 1 part");
  EXPECT_DEATH(
      directory_rename("cat/miaow/pig", "dog/woof"),
      "Cannot rename directory 'cat/miaow/pig' with 3 parts "
      "to 'dog/woof' with 2 parts");

  temp_filename_new(&_base, null);

  string_new_concat(&_dir1, _base->_filename, "/cat");
  directory_ensure_exists(_dir1);
  temp_filename_unlock(_base);
  string_new_concat(&_dir2, _base->_filename, "/dog");
  directory_ensure_exists(_dir2);
  string_new_f(
      &_e,
      "Directory '%s/' already exists, so cannot rename '%s/' to it",
      _dir2,
      _dir1);
  EXPECT_DEATH(directory_rename(_dir1, _dir2), _e);
  string_delete(&_e);
  directory_remove(_dir2);
  directory_remove(_dir1);
  string_delete(&_dir2);
  string_delete(&_dir1);

  string_new_concat(&_dir1, _base->_filename, "/cat/pig");
  directory_ensure_exists(_dir1);
  string_new_concat(&_dir2, _base->_filename, "/dog/pig");
  directory_ensure_exists(_dir2);
  string_new_f(
      &_e,
      "Directory '%s/dog/' already exists, so cannot rename '%s/cat/' to it",
      _base->_filename,
      _base->_filename);
  EXPECT_DEATH(directory_rename(_dir1, _dir2), _e);
  string_delete(&_e);
  directory_remove(_dir2);
  directory_remove(_dir1);
  string_delete(&_dir2);
  string_delete(&_dir1);

  string_new_concat(&_dir1, _base->_filename, "/woof/cat/pig");
  directory_ensure_exists(_dir1);
  string_new_concat(&_dir2, _base->_filename, "/woof/dog/pig");
  directory_ensure_exists(_dir2);
  string_new_f(
      &_e,
      "Directory '%s/woof/dog/' already exists, so cannot rename "
      "'%s/woof/cat/' to it",
      _base->_filename,
      _base->_filename);
  EXPECT_DEATH(directory_rename(_dir1, _dir2), _e);
  string_delete(&_e);
  directory_remove(_dir2);
  directory_remove(_dir1);
  string_delete(&_dir2);
  string_delete(&_dir1);

  temp_filename_delete(&_base);
}

FINISH_DEATH
